#include "App.h"

#include "AMI.h"
#include "AppState.h"
#include "DemoDataSource.h"
#include "DriverView.h"
#include "MissionSelect.h"
#include "events.h"

#include <SDL2/SDL.h>
#include <SDL2/SDL_events.h>
#include <SDL2/SDL_image.h>
#include <SDL2/SDL_keycode.h>
#include <SDL2/SDL_render.h>
#include <SDL2/SDL_timer.h>

#include <fmt/format.h>
#include <spdlog/spdlog.h>

#include <iostream>
#include <memory>
#include <queue>
#include <stdexcept>

App::App() {
  init_sdl();
  init_state();
}

App::~App() {
  // Destroy window
  SDL_DestroyRenderer(renderer);
  renderer = nullptr;
  SDL_DestroyWindow(window);
  window = nullptr;
}

void App::init_sdl() {
  window = SDL_CreateWindow("FT22 STW Display", SDL_WINDOWPOS_UNDEFINED,
                            SDL_WINDOWPOS_UNDEFINED, SCREEN_WIDTH,
                            SCREEN_HEIGHT, SDL_WINDOW_SHOWN);
  if (window == nullptr) {
    throw std::runtime_error(
        fmt::format("Couldn't create window: {}", SDL_GetError()));
  }
  // Software renderer is *MUCH* (~50x speedup) faster than hardware accelerated
  // renderer on a Pi Zero.
  renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_SOFTWARE);
  if (renderer == nullptr) {
    throw std::runtime_error(
        fmt::format("Couldn't create renderer: {}", SDL_GetError()));
  }
}

void App::init_state() {
  try {
    state = std::make_unique<AppState>("/dev/i2c-3");
  } catch (const std::runtime_error& e) {
    spdlog::error("Couldn't create AppState: {}", e.what());
  }
}

int App::run() {
  mission_select = std::make_unique<MissionSelect>(renderer);
  ami = std::make_unique<AMI>(renderer);
  driver_view = std::make_unique<DriverView>(renderer);

#ifndef NDEBUG
  std::cout << "Change brake balance with W (+) and S (-)" << std::endl;
#endif

  running = true;

  uint32_t second_start = SDL_GetTicks();
  unsigned frames = 0;
  while (running) {
    uint32_t now = SDL_GetTicks();
    if (now - second_start > 5000) {
      spdlog::info("{} FPS", frames / 5);
      second_start = now;
      frames = 0;
    }

    if (state != nullptr) {
      state->poll();
    }

    handle_events();
    render();

    frames++;
    SDL_Delay(1);
  }

  return 0;
}

void App::handle_events() {
  std::queue<Event> events;
  SDL_Event e;
  while (SDL_PollEvent(&e) != 0) {
    if (e.type == SDL_QUIT) {
      running = false;
#ifndef NDEBUG
    } else if (e.type == SDL_KEYUP) {
      keyboard_handler.handle_keyup(e, *state);
#endif // !NDEBUG
    }
  }
}

void App::render() {
  AppView view = state->get_view();
  switch (view) {
  case AppView::MISSION_SELECT:
    mission_select->draw(*state);
    break;
  case AppView::AMI:
    ami->draw(*state);
    break;
  case AppView::DRIVER:
    driver_view->draw(*state);
    break;
  default:
    throw std::runtime_error(fmt::format("Unknown view: {}", view));
  }
  SDL_RenderPresent(renderer);
}

SDLManager::SDLManager() {
  if (SDL_Init(SDL_INIT_VIDEO) < 0) {
    throw std::runtime_error(
        fmt::format("Couldn't initialize SDL: {}", SDL_GetError()));
  }
  if (TTF_Init() == -1) {
    throw std::runtime_error(
        fmt::format("Couldn't initialize SDL_ttf: {}", TTF_GetError()));
  }
  int img_flags = IMG_INIT_PNG;
  if (!(IMG_Init(img_flags) & img_flags)) {
    throw std::runtime_error(
        fmt::format("Couldn't initialitze SDL_image: {}", IMG_GetError()));
  }
}

SDLManager::~SDLManager() {
  IMG_Quit();
  TTF_Quit();
  SDL_Quit();
}

/*
Handle key presses
*/
void KeyboardHandler::handle_keyup(const SDL_Event& ev, AppState& state) {
  auto pressed_key = ev.key.keysym.sym;

  // we are in the first screen (mission select)
  if (state.view == AppView::MISSION_SELECT) {

    // handle change selected mission
    if (pressed_key == SDLK_DOWN || pressed_key == SDLK_RIGHT) {
      auto new_mission = static_cast<int>(state.mission) + 1;
      if (new_mission > 7) {
        new_mission = 1;
      }
      state.mission = static_cast<Mission>(new_mission);
    } else if (pressed_key == SDLK_UP || pressed_key == SDLK_LEFT) {
      auto new_mission = static_cast<int>(state.mission) - 1;
      if (new_mission < 1) {
        new_mission = 7;
      }
      state.mission = static_cast<Mission>(new_mission);

      // commit mission selection
    } else if (pressed_key == SDLK_RETURN) {

      if (state.mission == Mission::MANUAL) {
        state.view = AppView::DRIVER;
      } else {
        state.view = AppView::AMI;
      }
    }
  }

  // go back to mission select
  if (state.view == AppView::AMI || state.view == AppView::DRIVER) {
    if (pressed_key == SDLK_ESCAPE) {
      state.view = AppView::MISSION_SELECT;
    }
  }

  // handle mock brake balance change
  if (pressed_key == SDLK_w) {
    state.data_source.bump_brake_balance_up();
  } else if (pressed_key == SDLK_s) {
    state.data_source.bump_brake_balance_down();
  }
}
