#include "AppState.h"

#include <fmt/format.h>
#include <spdlog/spdlog.h>

#include <fcntl.h>
#include <optional>
extern "C" {
#include <i2c/smbus.h>
}
#include <linux/i2c-dev.h>
#include <sys/ioctl.h>

#include <cstdint>
#include <stdexcept>
#include <unistd.h>

std::ostream& operator<<(std::ostream& os, AppView view) {
  switch (view) {
  case AppView::MISSION_SELECT:
    return os << "MISSION_SELECT";
  case AppView::AMI:
    return os << "AMI";
  case AppView::DRIVER:
    return os << "DRIVER";
  case AppView::TESTING:
    return os << "TESTING";
  default:
    return os << "UNKNOWN_VIEW[" << static_cast<int>(view) << "]";
  }
}

std::ostream& operator<<(std::ostream& os, Mission mission) {
  switch (mission) {
  case Mission::NONE:
    return os << "NONE";
  case Mission::ACCELERATION:
    return os << "ACCELERATION";
  case Mission::SKIDPAD:
    return os << "SKIDPAD";
  case Mission::AUTOCROSS:
    return os << "AUTOCROSS";
  case Mission::TRACKDRIVE:
    return os << "TRACKDRIVE";
  case Mission::EBS_TEST:
    return os << "EBS_TEST";
  case Mission::INSPECTION:
    return os << "INSPECTION";
  case Mission::MANUAL:
    return os << "MANUAL";
  default:
    return os << "UNKNOWN_MISSION[" << static_cast<int>(mission) << "]";
  }
}

AppState::AppState(const std::string& i2c_dev_path)
    : view{AppView::MISSION_SELECT}, mission{Mission::ACCELERATION} {
  i2c_dev_file = open(i2c_dev_path.c_str(), O_RDWR);
  if (i2c_dev_file < 0) {
    spdlog::error("Couldn't open I2C device: {}", strerror(errno));
    i2c_dev_file = std::nullopt;
    return;
  }

  if (ioctl(*i2c_dev_file, I2C_SLAVE, I2C_SLAVE_ADDR) < 0) {
    spdlog::error("Couldn't set I2C slave address: {}", strerror(errno));
    close(*i2c_dev_file);
    i2c_dev_file = std::nullopt;
  }
}

AppState::~AppState() {
  if (i2c_dev_file) {
    close(*i2c_dev_file);
  }
}

void AppState::poll() {
  data_source.poll();
  if (!i2c_dev_file) {
    return;
  }

  uint8_t data[32];
  int count = i2c_smbus_read_block_data(*i2c_dev_file, I2C_POLL_COMMAND, data);
  if (count < 0) {
    throw std::runtime_error(
        fmt::format("Couldn't poll I2C slave: {}", strerror(errno)));
  }
  view = static_cast<AppView>(data[0]);
  spdlog::info("View after poll: {}", view);
  switch (view) {
  case AppView::MISSION_SELECT:
    unmarshal_mission_select(data);
    break;
  case AppView::AMI:
    unmarshal_ami(data);
    break;
  default:
    throw std::runtime_error(fmt::format("Unknown view: {}", view));
  }
}

AppView AppState::get_view() const { return view; }

Mission AppState::get_mission() const { return mission; }

void AppState::unmarshal_mission_select(uint8_t* data) {
  mission = static_cast<Mission>(data[1]);
  spdlog::info("Mission after poll: {}", mission);
}

void AppState::unmarshal_ami(uint8_t* data) {
  mission = static_cast<Mission>(data[1]);
  spdlog::info("Mission after poll: {}", mission);
}

DemoDataSource AppState::get_data_source() const { return data_source; }