#include "DriverView.h"

#include "DemoDataSource.h"
#include "defines.h"

#include <fmt/format.h>

#include <iostream>

#define LARGE_FONT_SIZE_POINTS 170
#define GIANT_FONT_SIZE_POINTS 250
#define MEDIUM_FONT_SIZE_POINTS 50
#define SOC_BAR_HEIGHT 80
#define SPEED_TEXT_X 22
#define SPEED_TEXT_Y 50

DriverView::DriverView(SDL_Renderer* renderer)
    : View{renderer}, font_large{util::load_font("resources/Avenir-Book.ttf",
                                                 LARGE_FONT_SIZE_POINTS)},
      font_detail{util::load_font("resources/Avenir-Book.ttf", 20)},
      font_tiny{util::load_font("resources/Avenir-Book.ttf", 15)},
      font_giant{
          util::load_font("resources/Avenir-Book.ttf", GIANT_FONT_SIZE_POINTS)},
      font_medium{util::load_font("resources/Avenir-Book.ttf",
                                  MEDIUM_FONT_SIZE_POINTS)} {
  speed_widget = std::make_unique<TextWidget>(renderer, font_large, "");
  speed_widget->set_position(SPEED_TEXT_X, SPEED_TEXT_Y);
  speed_widget->set_alignment(Alignment::CENTER, Alignment::TOP);

  speed_hint_widget = std::make_unique<TextWidget>(renderer, font_tiny, "SPD");
  speed_hint_widget->set_position(20, 210);
  speed_hint_widget->set_alignment(Alignment::CENTER, Alignment::TOP);

  brake_balance_widget = std::make_unique<TextWidget>(renderer, font_large, "");
  brake_balance_widget->set_position(SPEED_TEXT_X + SCREEN_WIDTH / 2,
                                     SPEED_TEXT_Y);
  brake_balance_widget->set_alignment(Alignment::CENTER, Alignment::TOP);
  brake_balance_hint_widget =
      std::make_unique<TextWidget>(renderer, font_tiny, "BB");
  brake_balance_hint_widget->set_position(20 + SCREEN_WIDTH / 2, 210);
  brake_balance_hint_widget->set_alignment(Alignment::CENTER, Alignment::TOP);

  general_info_widget = std::make_unique<TextWidget>(renderer, font_detail, "");
  general_info_widget->set_position(SPEED_TEXT_X, SPEED_TEXT_Y + 200);
  general_info_widget->set_alignment(Alignment::LEFT, Alignment::TOP);
  general_info_widget->set_wrap_width(SCREEN_WIDTH - 50);

  focus_widget = std::make_unique<TextWidget>(renderer, font_giant, "");
  focus_widget->set_position(120, 50);
  focus_widget->set_alignment(Alignment::LEFT, Alignment::TOP);
  focus_hint_widget = std::make_unique<TextWidget>(renderer, font_medium, "");
  focus_hint_widget->set_position(120, 30);
  // focus_hint_widget->set_alignment(Alignment::CENTER, Alignment::CENTER);
}

DriverView::~DriverView() {

  TTF_CloseFont(font_medium);
  TTF_CloseFont(font_large);
  TTF_CloseFont(font_detail);
  TTF_CloseFont(font_tiny);
  TTF_CloseFont(font_giant);
}

void DriverView::draw(const AppState& state) {
  auto ds = state.get_data_source();
  auto hv_ratio = ds.get_hv_voltage_ratio();
  SDL_Rect rect;
  rect.x = 0;
  rect.y = 0;
  rect.w = SCREEN_WIDTH;
  rect.h = SCREEN_HEIGHT;

  static auto hv_low = false;
  // hysteresis for the hv voltage low
  if (hv_ratio < 0.15) {
    hv_low = true;
  } else if (hv_ratio > 0.2) {
    hv_low = false;
  }

  int red = hv_low ? 0xa0 : 0x30;
  SDL_SetRenderDrawColor(renderer, red, 0x30, 0x30, 255);
  SDL_RenderFillRect(renderer, &rect);

  auto brake_balance_text = fmt::format("{:.0f}", ds.get_brake_balance() * 100);

  if (ds.brake_balance_change_is_recent()) {
    SDL_SetRenderDrawColor(renderer, 255, 69, 100, 255);
    SDL_RenderFillRect(renderer, &rect);
    focus_widget->update_text(brake_balance_text);
    focus_widget->draw();
    focus_hint_widget->update_text("BBAL");
    focus_hint_widget->draw();
    return;
  }

  rect.w = SCREEN_WIDTH * hv_ratio;
  rect.h = SOC_BAR_HEIGHT;
  SDL_SetRenderDrawColor(renderer, 76, 255, 0, 255);
  SDL_RenderFillRect(renderer, &rect);

  speed_widget->update_text(fmt::format("{:02.0f}", ds.get_speed()));
  speed_widget->draw();
  speed_hint_widget->draw();

  brake_balance_widget->update_text(brake_balance_text);

  brake_balance_widget->draw();
  brake_balance_hint_widget->draw();

  // draw rectangles around the speed and brake balance widgets
  SDL_SetRenderDrawColor(renderer, 0xFF, 0xFF, 0xFF, 255);
  rect.x = 0;
  rect.y = SOC_BAR_HEIGHT;
  rect.w = SCREEN_WIDTH / 2;
  rect.h = 150;
  SDL_RenderDrawRect(renderer, &rect);
  rect.x = SCREEN_WIDTH / 2;
  SDL_RenderDrawRect(renderer, &rect);

  auto general_string = fmt::format(
      "HV: {:.0f}V | LV: {:.1f}V | B_TEMP: {:.1f}C\nAPPS: {:.0f}% | BPF: "
      "{:.0f} bar | BPR: {:.0f} bar",
      ds.get_hv_voltage(), ds.get_lv_voltage(), ds.get_battery_temperature(),
      ds.get_throttle_ratio() * 100, ds.get_brake_pressure_front_bar(),
      ds.get_brake_pressure_rear_bar());

  general_info_widget->update_text(general_string);
  general_info_widget->draw();
}