#include "widgets.h"

#include "defines.h"
#include "util.h"

#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include <fmt/format.h>

#include <cstdint>
#include <stdexcept>

constexpr uint8_t LIST_SELECTION_BG_R = 0x77;
constexpr uint8_t LIST_SELECTION_BG_G = 0x33;
constexpr uint8_t LIST_SELECTION_BG_B = 0x33;
constexpr uint8_t LIST_OTHER_BG_R = 0x22;
constexpr uint8_t LIST_OTHER_BG_G = 0x22;
constexpr uint8_t LIST_OTHER_BG_B = 0x22;
constexpr uint8_t LIST_NORMAL_BG_R = 0x00;
constexpr uint8_t LIST_NORMAL_BG_G = 0x00;
constexpr uint8_t LIST_NORMAL_BG_B = 0x00;

std::ostream& operator<<(std::ostream& os, Alignment align) {
  switch (align) {
  case Alignment::START:
    return os << "START";
  case Alignment::CENTER:
    return os << "CENTER";
  case Alignment::END:
    return os << "END";
  default:
    return os << "UNKNOWN ALIGNMENT[" << static_cast<int>(align) << "]";
  }
}

PositionInfo::PositionInfo()
    : x{0}, y{0}, align_h{Alignment::LEFT}, align_v{Alignment::LEFT} {}

Widget::Widget(SDL_Renderer* renderer) : renderer{renderer}, rect{0, 0, 0, 0} {}

Widget::~Widget() {}

void Widget::set_width(int width, bool preserve_aspect_ratio) {
  if (preserve_aspect_ratio) {
    float scale = ((float)width) / rect.w;
    rect.h = round(rect.h * scale);
  }
  rect.w = width;
  recalculate_pos();
}

void Widget::set_height(int height, bool preserve_aspect_ratio) {
  if (preserve_aspect_ratio) {
    float scale = ((float)height) / rect.h;
    rect.w = round(rect.w * scale);
    recalculate_pos();
  }
  rect.h = height;
}

void Widget::set_position(int x, int y) {
  pos.x = x;
  pos.y = y;
  recalculate_pos();
}

void Widget::set_alignment(Alignment align_h, Alignment align_v) {
  pos.align_h = align_h;
  pos.align_v = align_v;
  recalculate_pos();
}

int Widget::get_width() { return rect.w; }

int Widget::get_height() { return rect.h; }

const PositionInfo& Widget::get_position() { return pos; }

void Widget::recalculate_pos() {
  int x = pos.x;
  int y = pos.y;

  switch (pos.align_h) {
  case Alignment::START:
    break;
  case Alignment::END:
    x -= rect.w;
    break;
  case Alignment::CENTER:
    x -= rect.w / 2;
    break;
  default:
    throw std::runtime_error(fmt::format("Unknown alignment: {}", pos.align_h));
  }
  switch (pos.align_v) {
  case Alignment::START:
    break;
  case Alignment::END:
    y -= rect.h;
    break;
  case Alignment::CENTER:
    y -= rect.h / 2;
    break;
  default:
    throw std::runtime_error(fmt::format("Unknown alignment: {}", pos.align_v));
  }

  rect.x = x;
  rect.y = y;
}

TextureWidget::TextureWidget(SDL_Renderer* renderer,
                             std::optional<SDL_Texture*> texture)
    : Widget{renderer}, texture{texture} {
  if (texture) {
    update_dimensions_from_texture(*texture);
  }
}

TextureWidget::~TextureWidget() {
  if (texture) {
    SDL_DestroyTexture(*texture);
    texture = nullptr;
  }
}

void TextureWidget::draw() {
  if (texture) {
    SDL_RenderCopy(renderer, *texture, NULL, &rect);
  }
}

void TextureWidget::update_texture(SDL_Texture* new_texture) {
  if (texture) {
    SDL_DestroyTexture(*texture);
  }
  update_dimensions_from_texture(new_texture);
  texture = new_texture;
}

void TextureWidget::update_dimensions_from_texture(SDL_Texture* texture) {
  if (SDL_QueryTexture(texture, nullptr, nullptr, &rect.w, &rect.h) != 0) {
    throw std::runtime_error(
        fmt::format("Couldn't query texture: {}", SDL_GetError()));
  }
}

ImageWidget::ImageWidget(SDL_Renderer* renderer, const std::string& path)
    : TextureWidget{renderer, util::load_img(renderer, path)} {}

TextWidget::TextWidget(SDL_Renderer* renderer, TTF_Font* font,
                       const std::string& initial_text)
    : TextureWidget{renderer, std::nullopt}, font{font}, text{initial_text} {
  if (text != "") {
    update_texture(generate_text(text));
  }
}

TextWidget::~TextWidget() {}

void TextWidget::update_text(const std::string& new_text) {
  if (text != new_text) {
    update_texture(generate_text(new_text));
    text = new_text;
  }
}

SDL_Texture* TextWidget::generate_text(const std::string& text) {
  SDL_Surface* surf;
  if (wrap_width == -1) {
    surf = TTF_RenderText_Blended(font, text.c_str(), {0xFF, 0xFF, 0XFF, 0xFF});
  } else {
    surf = TTF_RenderText_Blended_Wrapped(font, text.c_str(),
                                          {0xFF, 0xFF, 0XFF, 0xFF}, wrap_width);
  }
  if (surf == nullptr) {
    throw std::runtime_error(
        fmt::format("Unable to render text surface: {}", TTF_GetError()));
  }
  SDL_Texture* texture = SDL_CreateTextureFromSurface(renderer, surf);
  if (texture == nullptr) {
    throw std::runtime_error(fmt::format(
        "Unable to create texture from text surface: {}", SDL_GetError()));
  }
  return texture;
}

void TextWidget::set_wrap_width(int width) { wrap_width = width; }

ListWidget::ListWidget(SDL_Renderer* renderer, int element_height,
                       Alignment element_alignment)
    : Widget{renderer}, element_height{element_height},
      element_alignment{element_alignment}, selection{0} {}

ListWidget::~ListWidget() {}

void ListWidget::add_element(Widget* element) {
  int new_index = elements.size();
  elements.push_back(element);
  place_element(element, new_index);
}

void ListWidget::draw() {
  int max_y = rect.y + rect.h;
  for (size_t i = 0; i < elements.size(); i++) {
    const auto& element = elements[i];
    const PositionInfo& element_pos = element->get_position();
    if (element_pos.y >= max_y) {
      // Since the elements are sorted, we can stop the loop now.
      break;
    }
    if (i == selection) {
      SDL_SetRenderDrawColor(renderer, LIST_SELECTION_BG_R, LIST_SELECTION_BG_G,
                             LIST_SELECTION_BG_B, 0xFF);
    } else if (i % 2 == 1) {
      SDL_SetRenderDrawColor(renderer, LIST_OTHER_BG_R, LIST_OTHER_BG_G,
                             LIST_OTHER_BG_B, 0xFF);
    } else {
      SDL_SetRenderDrawColor(renderer, LIST_NORMAL_BG_R, LIST_NORMAL_BG_G,
                             LIST_NORMAL_BG_B, 0xFF);
    }
    SDL_Rect fill_rect = {
        .x = pos.x, .y = element_pos.y, .w = rect.w, .h = element_height};
    SDL_RenderFillRect(renderer, &fill_rect);
    element->draw();
    if (i != elements.size() - 1) {
      int border_y = element_pos.y + element->get_height();
      SDL_SetRenderDrawColor(renderer, 0xFF, 0xFF, 0xFF, 0xFF);
      SDL_RenderDrawLine(renderer, rect.x, border_y, rect.x + rect.w, border_y);
    }
  }
}

void ListWidget::select_next() {
  size_t n = elements.size();
  if (n == 0) {
    return;
  }
  selection = (selection + 1) % n;
}

void ListWidget::select_prev() {
  size_t n = elements.size();
  if (n == 0) {
    return;
  }
  if (selection == 0) {
    selection = n - 1;
  } else {
    selection = selection - 1;
  }
}

void ListWidget::select(size_t n) {
  if (n > elements.size()) {
    throw std::runtime_error(fmt::format(
        "Tried to select element {}, but there are only {} elements!", n,
        elements.size()));
  }
  selection = n;
}

size_t ListWidget::get_selection() { return selection; }

void ListWidget::place_element(Widget* element, int index) {
  int x = rect.x;
  switch (element_alignment) {
  case Alignment::LEFT:
    break;
  case Alignment::RIGHT:
    x += rect.w;
    break;
  case Alignment::CENTER:
    x += rect.w / 2;
    break;
  default:
    throw std::runtime_error(
        fmt::format("Unknown alignment: {}", element_alignment));
  }
  // Additional pixel for border
  int y = rect.y + index * (element_height + 1);
  element->set_position(x, y);
  element->set_alignment(element_alignment, Alignment::TOP);
  element->set_height(element_height);
}
